from collections import abc
from functools import cache
from itertools import repeat
import logging
from pathlib import Path
import re
from typing import Any, NamedTuple, TypeAlias

from catgram.category_tree import CategoryTree
from catgram.ccg.candc_ignore import ignore_dep
from catgram.ccg.heads import HeadDependency, WordToken


logger = logging.getLogger("catgram.ccg.dependencies")


class DependencyEdge(NamedTuple):
    pred: WordToken
    arg: WordToken


class DependencyLabel(NamedTuple):
    cat: CategoryTree
    slot: int


class DependencyScores(NamedTuple):
    n_deps_tru: int
    n_deps_hyp: int
    correct: int
    precision: float
    recall: float
    f1: float


DependencySet: TypeAlias = dict[DependencyEdge, DependencyLabel]


class _LevState(NamedTuple):
    distance: int
    back: tuple[tuple[int, int], ...]
    is_match: bool


@cache
def _lev_matches(
    tru: abc.Sequence[str], hyp: abc.Sequence[str]
) -> frozenset[tuple[int, int]]:
    tru = (*tru, "<end>")
    hyp = (*hyp, "<end>")
    # numpy would be convenient, but i'd prefer to avoid the additional
    # dependency just for this evaluation script; plus this way i can store
    # backpointers and match status for each position
    d = {}
    # technically there should be backpointers in the first row and column but
    # since the ultimate interest is in optimal match states, this row/column
    # will just be ignored for the backward pass
    for i in range(len(tru) + 1):
        d[i, 0] = _LevState(i, (), False)
    for j in range(1, len(hyp) + 1):
        d[0, j] = _LevState(j, (), False)
    for i, t in enumerate(tru, start=1):
        for j, h in enumerate(hyp, start=1):
            sub = 0 if t == h else 1
            dist = min(
                d_del := d[i - 1, j].distance + 1,
                d_ins := d[i, j - 1].distance + 1,
                d_sub := d[i - 1, j - 1].distance + sub,
            )
            back = []
            if i > 1 and j > 1:
                if dist == d_del:
                    back.append((i - 1, j))
                if dist == d_ins:
                    back.append((i, j - 1))
                if dist == d_sub:
                    back.append((i - 1, j - 1))
            d[i, j] = _LevState(dist, tuple(back), sub == 0)

    # and now the backward pass
    lev_matches = set()
    stack = list(d[len(tru), len(hyp)].back)
    while stack:
        i, j = stack.pop()
        state = d[i, j]
        if state.is_match:
            lev_matches.add((i - 1, j - 1))
        stack.extend(state.back)
    return frozenset(lev_matches)


def dep_label_correct(
    tru_label: DependencyLabel,
    hyp_label: DependencyLabel,
    subcat_label: bool,
    subcat_align: bool,
) -> bool:
    """Determine whether the label for a single dependency is correct.

    Note that label correctness is a symmetric measure, but the convention is to
    consider the first set of dependencies to be ground truth so that the
    score meanings are clear.

    Args:
        tru_label (DependencyLabel): The ground-truth dependency label.
        hyp_label (DependencyLabel): The predicted dependency label.
        subcat_label (bool): Whether to use subcategorial labelling as defined
            by Bhargava and Penn (ACL 2023) as opposed to standard,
            full-category labelling.
        subcat_align (bool): Whether to use subcategorial alignment as defined
            by Bhargava and Penn (ACL 2023). Has no effect if `subcat_label` is
            False.

    Returns:
        bool: True if the label is correct according to the specified scoring
            attributes, False otherwise.
    """
    if not subcat_label:
        # no subcategorial labelling, so the full label labels must be equal
        return tru_label == hyp_label
    tru_subcat = tru_label.cat.get_arg(tru_label.slot)
    hyp_subcat = hyp_label.cat.get_arg(hyp_label.slot)
    if tru_subcat != hyp_subcat:
        # if the subcategorial label is incorrect, there's no recovering...
        return False
    # subcategorial label is correct
    if tru_label.slot == hyp_label.slot:
        # slot is correct, no need to think about alignment
        return True
    # subcategorial label is correct but slot is incorrect
    elif not subcat_align:
        # no alignment to accommodate incorrect slot, so label is incorrect
        return False
    # subcategorial label is correct but slot is incorrect; try to find
    # plausible alignment
    tru_fs, hyp_fs = tru_label.cat.func_seq, hyp_label.cat.func_seq
    return (tru_label.slot, hyp_label.slot) in _lev_matches(tru_fs, hyp_fs)


def score_sentence(
    tru_deps: DependencySet,
    hyp_deps: DependencySet,
    *,
    labelling: str | None = "subcat",
    subcat_align: bool = True,
    ignore_root: bool = False,
    sentence_id: Any = None,
) -> DependencyScores:
    """Score the dependencies of a single sentence.

    Note that the score is a symmetric measure, but the convention is to
    consider the first set of dependencies to be ground truth so that the
    meaning of precision and recall is clear.

    Args:
        tru_deps (DependencySet): The ground-truth dependency set.
        hyp_deps (DependencySet): The predicted dependency set.
        labelling (str | None, optional): The labelling method. Can be one of
            'subcat', 'full', or None. 'subcat' is subcategorial labelling as
            defined by Bhargava and Penn (ACL 2023), 'full' is standard,
            full-category labelling as in the standard CCG evaluation, and
            None is no labelling (for computing unlabelled scores). Defaults to
            "subcat".
        subcat_align (bool, optional): Whether to use subcategorial alignment
            as defined by Bhargava and Penn (ACL 2023). Has no effect if
            `labelling` is not 'subcat'. Defaults to True.
        ignore_root (bool, optional): Ignore root dependencies that are in the
            dependency sets and supress warnings about missing root
            dependencies. Defaults to False.
        sentence_id (Any, optional): A label for this sentence. Only used for
            identifying the sentence when issuing warnings. Defaults to None.

    Raises:
        ValueError: If `labelling` is not one of 'subcat', 'full', or None.

    Returns:
        DependencyScores: A named tuple containing relevant scores and counts
            for the given pair of dependency sets. The tuple includes precision,
            recall, and F1, though beware that when evaluating on a corpus of
            sentences, they should not be used as the standard evaluation is
            a micro-average (i.e., computed from aggregated counts rather than
            averaging per-sentence scores).
    """
    if ignore_root:
        tru_deps = {k: v for k, v in tru_deps.items() if k.pred != ("ROOT", 0)}
        hyp_deps = {k: v for k, v in hyp_deps.items() if k.pred != ("ROOT", 0)}
    else:
        if sentence_id is None or sentence_id == "":
            sent_suffix = ""
        else:
            sent_suffix = f" for sentence {sentence_id}"
        if tru_deps and ("ROOT", 0) not in [edge[0] for edge in tru_deps]:
            logger.warning(f"No root in ground-truth dependencies{sent_suffix}")
        if hyp_deps and ("ROOT", 0) not in [edge[0] for edge in hyp_deps]:
            logger.warning(f"No root in predicted dependencies{sent_suffix}")
    n_tru, n_hyp = len(tru_deps), len(hyp_deps)
    if (n_tru == n_hyp == 0) or (
        n_tru == n_hyp == 1
        and list(tru_deps)[0].pred == list(hyp_deps)[0].pred == ("ROOT", 0)
    ):
        # if there are no (non-ROOT) dependencies, score like the standard CCG
        # evaluation (zero contribution to micro-averaged stats); but it doesn't
        # make sense to say precision/recall/F1 are zero for such sentences
        # the most important thing is that the scores are consistent between the
        # case with roots and the case without so that, when compared, the
        # difference doesn't result in treating the case as a rank inversion
        return DependencyScores(
            n_deps_tru=0, n_deps_hyp=0, correct=0, precision=1.0, recall=1.0, f1=1.0
        )
    tru_edges, hyp_edges = set(tru_deps), set(hyp_deps)
    correct_edges = tru_edges & hyp_edges  # start from unlabelled F1
    match labelling:
        case "full" | "subcat":
            correct_edges = {
                e
                for e in correct_edges
                if dep_label_correct(
                    tru_deps[e], hyp_deps[e], labelling == "subcat", subcat_align
                )
            }
        case None:
            pass
        case _:
            raise ValueError(
                "`labelling` must be one of None, 'full', or 'subcat'; "
                f"received {labelling:r}"
            )

    n_correct = len(correct_edges)
    return DependencyScores(
        n_deps_tru=n_tru,
        n_deps_hyp=n_hyp,
        correct=n_correct,
        precision=n_correct / n_hyp if n_hyp else 0.0,
        recall=n_correct / n_tru if n_tru else 0.0,
        f1=2 * n_correct / (n_tru + n_hyp),
    )


def file_deps(
    fn: Path | str, roots: abc.Iterator[HeadDependency | None] | None = None
) -> abc.Generator[DependencySet, None, None]:
    """Read CCG dependencies from a file.

    Args:
        fn (Path | str): The name of the file to read. Accepted formats are
            .deps files produced by Java C&C or C&C's `generate` program (run on
            .auto files produced by a system), CCGbank .parg files, and
            .ccgbank_deps files produced by the `parg2ccgbank_deps` script from
            C&C.
        roots (abc.Iterator[HeadDependency | None] | None, optional): An
            iterator of root dependencies to add to each sentence's dependency
            set. If None, no root dependencies are added. Defaults to None.

    Raises:
        ValueError: If the (optional) commented lines at top of the file aren't
            separated from the rest of the file by a blank line.
        ValueError: If a line has an unexpected number of fields.

    Yields:
        DependencySet: The dependency set for each sentence in the file.
    """
    with open(fn) as f:
        # most files have some comments at the top (indicating the command that
        # produced them, etc.), so test for those while still allowing for
        # a file without them
        if (pre := f.read(2))[0] == "#":
            f.seek(0)
            while line := next(f).strip():
                if line[0] != "#":
                    raise ValueError(
                        f"Unexpected uncommented line at top of {fn}: {line}"
                    )
        else:
            f.seek(0)
        parg = pre == "<s"

        sent_deps = {}
        roots_ = repeat(None) if roots is None else roots
        for line in f:
            if line[:3] == "<c>":
                # skip C&C's POS/category tag line
                continue
            elif parg and line[:2] == "<s":
                # reading CCGbank PARG file, skip sentence start line
                continue
            fields = line.strip().split()
            if not fields or fields[0] == "<\\s>":
                try:
                    root = next(roots_)
                except StopIteration:
                    raise ValueError(
                        "Not enough roots provided for sentences"
                    ) from None
                if root is not None:
                    root_token, root_cat = root
                    sent_deps[
                        DependencyEdge(WordToken("ROOT", 0), root_token)
                    ] = DependencyLabel(root_cat, 0)
                yield sent_deps
                sent_deps = {}
                continue

            if parg and 5 < len(fields) < 8:
                # direct from CCGbank PARG files
                arg_idx, pred_idx, cat, slot, arg, pred = fields[:6]
                pred = WordToken(pred, int(pred_idx) + 1)
                arg = WordToken(arg, int(arg_idx) + 1)
                # pred_idx = int(pred_idx) + 1
                # arg_idx = int(arg_idx) + 1
            else:
                if 4 < len(fields) < 7:
                    # from direct system outputs (e.g., Java C&C) or using C&C's
                    # `generate` program on `.auto` file produced by system
                    pred, cat, slot, arg, rule = fields[:5]
                    if ignore_dep(
                        pred.split("_")[0], cat, slot, arg.split("_")[0], rule
                    ):
                        continue
                    # remove the markup
                    cat = re.sub(r"<[0-9]>|\{[A-Z_]\*?\}|\[X\]", "", cat)[1:-1]
                elif len(fields) == 4:
                    # from CCGbank PARG files that were converted to ccgbank_deps
                    # files using the `parg2ccgbank_deps` script from C&C
                    pred, cat, slot, arg = fields
                else:
                    raise ValueError(f"Unexpected number of fields in line: {line}")
                pred = WordToken.from_str(pred)
                arg = WordToken.from_str(arg)
                # pred, pred_idx = pred.split("_")
                # arg, arg_idx = arg.split("_")
                # pred_idx = int(pred_idx)
                # arg_idx = int(arg_idx)

            # pred = WordToken(pred, pred_idx)
            # arg = WordToken(arg, arg_idx)
            cat = CategoryTree.from_str(cat)
            slot = int(slot)
            sent_deps[DependencyEdge(pred, arg)] = DependencyLabel(cat, slot)
    if roots is not None:
        try:
            next(roots_)
        except StopIteration:
            pass
        else:
            raise ValueError("More roots provided than sentences")



#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
from sys import stdin

from catgram.ccg.heads import head_methods, parse_roots


def main():
    parser = ArgumentParser(
        description="Extract roots from CCGbank-style .auto files",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "auto",
        nargs="?",
        type=FileType("r"),
        default=stdin,
        help=".auto file from which to extract heads",
    )
    # TODO: if a new head-finding method is added, the descriptions of each
    #  should go somewhere centralized so that the text can be used here
    #  and elsewhere as needed while staying consistent. an alternative option
    #  would be to put the add_argument call in a dedicated function that
    #  can then be imported elsewhere as needed (assuming the only use is
    #  for CLI param parsing)
    parser.add_argument(
        "--method",
        "-m",
        choices=head_methods,
        default=head_methods[0],
        help="head-finding method; 'autofile' follows the head specified by "
        "the .auto tree format, which is only applicable for CCGbank ground "
        "truth or EasyCCG predictions; 'ls14' follows the head-finding rules "
        "of Lewis and Steedman (EMNLP 2014; i.e., those of EasyCCG) for use "
        "with other parsers",
    )
    parser.add_argument(
        "--autox",
        "-x",
        action="store_true",
        help="force treatment of input file as an .autox file (as produced by "
        "DepCCG)",
    )

    args = parser.parse_args()

    if args.auto.name.endswith(".autox"):
        args.autox = True
    for head_dep in parse_roots(args.auto, args.method, args.autox):
        print(head_dep)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from collections import abc
import csv
from pathlib import Path
from sys import stdout

from catgram.ccg.dependencies import DependencyScores, file_deps, score_sentence
from catgram.ccg.heads import HeadDependency, parse_roots, head_methods


def _roots_iter(
    file: Path, method: str
) -> abc.Generator[HeadDependency | None, None, None]:
    if file.name.endswith(".roots"):
        with open(file) as f:
            for line in f:
                yield HeadDependency.from_str(line.strip())
    else:
        yield from parse_roots(file, method, file.name.endswith(".autox"))


def main():
    parser = ArgumentParser(
        description="Scoring of CCGbank-style dependencies",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "truth",
        type=Path,
        help="ground truth dependencies file",
    )
    parser.add_argument(
        "eval",
        type=Path,
        help="dependencies file to evaluate",
    )
    parser.add_argument(
        "truth_roots",
        nargs="?",
        type=Path,
        help="ground-truth roots file or .auto file to extract roots from",
    )
    parser.add_argument(
        "eval_roots",
        nargs="?",
        type=Path,
        help="evaluation roots file or .auto file to extract roots from",
    )
    parser.add_argument(
        "--no-subcat",
        "-c",
        action="store_false",
        dest="subcat",
        help="disable subcategorial labelling (and alignment)",
    )
    parser.add_argument(
        "--no-subcat-align",
        "-a",
        action="store_false",
        dest="subcat_align",
        help="disable subcategorial alignment",
    )
    parser.add_argument(
        "--root-method",
        "-m",
        choices=head_methods,
        default=head_methods[1],
        help="head-finding method for predicted dependencies; "
        f"'{head_methods[0]}' follows the head specified by the .auto tree "
        "format, which is only applicable for CCGbank ground truth or EasyCCG "
        f"predictions; '{head_methods[1]}' follows the head-finding rules of "
        "Lewis and Steedman (EMNLP 2014; i.e., those of EasyCCG) for use with "
        "other parsers; ignored if reading from .roots file",
    )
    parser.add_argument(
        "--ignore-roots",
        "-r",
        action="store_true",
        help="ignore root dependencies if present in the input and do not warn"
        " about their absence if not",
    )
    parser.add_argument(
        "--std",
        "-s",
        action="store_true",
        help="use original CCGbank evaluation; equivalent to -car",
    )
    parser.add_argument(
        "--unlabelled",
        "-u",
        action="store_true",
        help="use unlabelled scoring",
    )
    parser.add_argument(
        "--tsv",
        "-t",
        action="store_true",
        help="output in TSV format",
    )
    parser.add_argument(
        "--each-sentence",
        "-e",
        action="store_true",
        help="output metrics for each sentence; implies -t",
    )

    args = parser.parse_args()
    if args.std:
        args.subcat = args.subcat_align = False
        args.ignore_roots = True
    if args.unlabelled:
        labelling = None
    else:
        labelling = "subcat" if args.subcat else "full"
    if args.each_sentence:
        args.tsv = True

    n_correct = n_tru = n_hyp = 0

    if args.truth_roots:
        args.truth_roots = _roots_iter(args.truth_roots, head_methods[0])
    if args.eval_roots:
        args.eval_roots = _roots_iter(args.eval_roots, args.root_method)
    if args.tsv:
        writer = csv.writer(stdout, delimiter="\t", lineterminator="\n")
        if args.each_sentence:
            writer.writerow(["sent_id", *DependencyScores._fields])
        else:
            writer.writerow(DependencyScores._fields)

    i = 1
    for tru_deps, hyp_deps in zip(
        file_deps(args.truth, args.truth_roots),
        file_deps(args.eval, args.eval_roots),
        strict=True,
    ):
        scores = score_sentence(
            tru_deps,
            hyp_deps,
            labelling=labelling,
            subcat_align=args.subcat_align,
            ignore_root=args.ignore_roots,
            sentence_id=f"#{i}",
        )
        if args.each_sentence:
            writer.writerow([i, *scores])
        i += 1
        n_tru += scores.n_deps_tru
        n_hyp += scores.n_deps_hyp
        n_correct += scores.correct

    if n_tru == n_hyp == 0:
        precision = recall = f1 = 1.0
    else:
        precision = n_correct / n_hyp if n_hyp else 0.0
        recall = n_correct / n_tru if n_tru else 0.0
        f1 = 2 * n_correct / (n_tru + n_hyp)
    if not args.each_sentence:
        if args.tsv:
            writer.writerow([n_tru, n_hyp, n_correct, precision, recall, f1])
        else:
            print(f"Precision: {precision:.1%}\nRecall: {recall:.1%}\nF1: {f1:.1%}")


if __name__ == "__main__":
    main()



from collections import abc
from contextlib import nullcontext
from functools import partial
from pathlib import Path
import re
from string import punctuation
from typing import NamedTuple, Optional

from catgram.category_tree import CategoryTree


punctuation = set(punctuation) - set(r"\/()[]")
node_re = re.compile(r"\(<([TL]) ([^ >]+) ([^ >]+)(?: ([^ >]+?)){,4} ([^ >]+?)>")
brackets_re = re.compile(r"[LR][RC]B")
feat_re = re.compile(r"\[[a-z]+]")
slash_re = re.compile(r"\\/")


class WordToken(NamedTuple):
    # "token" here means unique occurence of a token in the sentence (as in the
    # type-token distinction), *not* the new-school meaning of "piece that goes
    # into neural language model"
    word: str
    idx: int

    def __str__(self) -> str:
        return "_".join(map(str, self))

    @classmethod
    def from_str(cls, s: str) -> "WordToken":
        word, idx = s.split("_")
        return cls(word, int(idx))


class HeadDependency(NamedTuple):
    token: WordToken
    cat: CategoryTree

    def __str__(self) -> str:
        return " ".join(map(str, self))

    @classmethod
    def from_str(cls, s: str) -> Optional["HeadDependency"]:
        if s == str(None):
            return None
        token, cat = s.split(" ")
        return cls(WordToken.from_str(token), CategoryTree.from_str(cat))


class _MalformedAutoTreeError(Exception):
    pass


def _head_autofile(
    it: abc.Iterator[re.Match], idx: int = 0, head_idx: int = 3, word_idx: int = 4
) -> tuple[HeadDependency, int]:
    # just follow the head specified by the relevant field in CCGbank's .auto
    # format. AFAIK only CCGbank uses this field for a semantic head; most
    # parsers just use it for the syntactic head that can otherwise be easily
    # determined from the input and output functors (at least for the
    # CCG-proper rules). the main exception is EasyCCG, which is the parser
    # associated with Lewis and Steedman (2014), which follows the head-finding
    # rules specified in that paper and implemented in _head_ls14
    m = next(it)
    node_cat = CategoryTree.from_str(re.sub(r"\[conj]$", "", m[2]))
    match m[1]:
        case "T":
            (head_l, _), idx = _head_autofile(it, idx)
            if m[5] == "2":
                (head_r, _), idx = _head_autofile(it, idx)
                return HeadDependency([head_l, head_r][int(m[head_idx])], node_cat), idx
            elif m[5] == "1":
                return HeadDependency(head_l, node_cat), idx
        case "L":
            idx += 1
            return (
                HeadDependency(
                    WordToken(slash_re.sub("/", m[word_idx]), idx), node_cat
                ),
                idx,
            )
    raise _MalformedAutoTreeError()


def _head_ls14(
    it: abc.Iterator[re.Match], idx: int = 0, head_idx: int = 3, word_idx: int = 4
) -> tuple[HeadDependency, int]:
    m = next(it)
    node_cat_str = re.sub(r"\[X]", "", m[2])  # for C&C
    node_cat = CategoryTree.from_str(node_cat_str)
    match m[1]:
        case "T":
            (head_l, cat_l), idx = _head_ls14(it, idx, head_idx, word_idx)
            if m[5] == "2":
                (head_r, cat_r), idx = _head_ls14(it, idx, head_idx, word_idx)
                cat_l_s, cat_r_s = str(cat_l), str(cat_r)
                if "conj" in node_cat_str and (
                    "conj" not in cat_l_s or "conj" not in cat_r_s
                ):
                    raise ValueError("'conj' parent nodes must have 'conj' children")
                if set(cat_l_s) & punctuation or brackets_re.match(cat_l_s):
                    if set(cat_r_s) & punctuation or brackets_re.match(cat_r_s):
                        # when both children are punctuation categories, choose
                        # the one that matches the parent as head, preferring
                        # the left child if both do
                        if node_cat_str == cat_l_s:
                            return HeadDependency(head_l, node_cat), idx
                        if node_cat_str == cat_r_s:
                            return HeadDependency(head_r, node_cat), idx
                        raise ValueError(
                            "one child must project to parent when both "
                            "children are punctuation categories"
                        )
                    # return non-punctuation child as head
                    return HeadDependency(head_r, node_cat), idx
                if set(cat_r_s) & punctuation or brackets_re.match(cat_r_s):
                    # return non-punctuation child as head
                    return HeadDependency(head_l, node_cat), idx
                if "conj" in cat_l_s:
                    if "conj" in cat_r_s:
                        if "conj" not in node_cat_str:
                            raise ValueError(
                                "two 'conj' children must combine to a 'conj' parent"
                            )
                    else:
                        # return non-conj child as head
                        return HeadDependency(head_r, node_cat), idx
                if "conj" in cat_r_s and "conj" not in cat_l_s:
                    raise ValueError(
                        "right child can be 'conj' only when left and parent "
                        "are also 'conj'"
                    )

                # for non-punct/conj cases, default to setting the head to the
                # principal/primary functor, but use the subordinate/secondary
                # functor if the principal looks like adjunction or like it was
                # produced by forward type-raising
                node_nf = CategoryTree.from_str(feat_re.sub("", node_cat_str))
                catl_nf = CategoryTree.from_str(feat_re.sub("", cat_l_s))
                catr_nf = CategoryTree.from_str(feat_re.sub("", cat_r_s))
                if catl_nf.root == "/":
                    if (
                        (catl_nf.result == node_nf and catl_nf.argument == catr_nf)
                        or (
                            catr_nf.root == node_nf.root == "/"
                            and catl_nf.result == node_nf.result
                            and catr_nf.argument == node_nf.argument
                            and catl_nf.argument == catr_nf.result
                        )
                        or (
                            catr_nf.root == node_nf.root
                            and node_nf.result is not None
                            and catr_nf.result.root == node_nf.result.root == "/"
                            and catl_nf.result == node_nf.result.result
                            and catr_nf.argument == node_nf.argument
                            and catr_nf.result.argument == node_nf.result.argument
                            and catl_nf.argument == catr_nf.result.result
                        )
                    ):
                        # A/ or B/ or B2/ or xB2/
                        if (cat_l.result == cat_l.argument) or (
                            cat_l.argument.root == "\\"
                            and cat_l.result == cat_l.argument.result
                        ):
                            # X/X or X/(X\Y)
                            return HeadDependency(head_r, node_cat), idx
                        return HeadDependency(head_l, node_cat), idx
                    if (
                        node_nf.root == catl_nf.root == catr_nf.root == "/"
                        and node_nf.argument == catl_nf.argument == catr_nf.argument
                        and catr_nf.result.root == "\\"
                        and catr_nf.result.argument == catl_nf.result
                        and catr_nf.result.result == node_nf.result
                    ):
                        # xS\
                        return HeadDependency(head_r, node_cat), idx
                if catr_nf.root == "\\":
                    if (
                        (catr_nf.result == node_nf and catr_nf.argument == catl_nf)
                        or (
                            catl_nf.root == node_nf.root
                            and node_nf.result is not None
                            and catr_nf.result == node_nf.result
                            and catl_nf.argument == node_nf.argument
                            and catr_nf.argument == catl_nf.result
                        )
                        or (
                            catl_nf.root == node_nf.root
                            and node_nf.result is not None
                            and catl_nf.result.root == node_nf.result.root == "/"
                            and catr_nf.result == node_nf.result.result
                            and catl_nf.argument == node_nf.argument
                            and catl_nf.result.argument == node_nf.result.argument
                            and catr_nf.argument == catl_nf.result.result
                        )
                    ):
                        # A\ or B\ or xB\ or xB2\
                        if (cat_r.result == cat_r.argument) or (
                            cat_r.argument.root == "/"
                            and cat_r.result == cat_r.argument.result
                        ):
                            # X\X or X\(X/Y)
                            return HeadDependency(head_l, node_cat), idx
                        return HeadDependency(head_r, node_cat), idx
            elif m[5] == "1":
                return HeadDependency(head_l, node_cat), idx
        case "L":
            idx += 1
            return (
                HeadDependency(
                    WordToken(slash_re.sub("/", m[word_idx]), idx), node_cat
                ),
                idx,
            )
    raise _MalformedAutoTreeError()


def parse_head(auto: str, method: str, *, autox: bool = False) -> HeadDependency:
    """
    Get the head of a node in a tree in CCGbank's .auto format (including
    DepCCG's .autox variation).

    Args:
        auto (str): The .auto tree node to find the head of.
        method (str): The head-finding method to use. Options are 'autofile',
            which follows the head specification syntax as defined by CCGbank,
            and 'ls14', which follows the head-finding algorithm rules of
            Lewis and Steedman (EMNLP 2014; i.e., those of EasyCCG) for use with
            other parsers.
        autox (bool, optional): Whether `auto` is in DepCCG's .autox format.
            Defaults to False.

    Raises:
        ValueError: If `method` is not a valid head-finding method.
        ValueError: If `auto` is malformed.
        ValueError: If the head-finding method throws an error.

    Returns:
        HeadDependency: The head of the node in `auto` as found by `method`.
    """
    try:
        _parse_head = globals()[f"_head_{method}"]
    except KeyError:
        raise ValueError(f"Unknown head-finding method: {method!r}") from None
    if autox:
        _parse_head = partial(_parse_head, head_idx=4, word_idx=3)

    try:
        head, _ = _parse_head(it := node_re.finditer(auto))
    except _MalformedAutoTreeError:
        raise ValueError(f"Malformed .auto tree: {auto!r}") from None
    except Exception:
        raise ValueError(f"Error in parsing .auto tree: {auto!r}")
    try:
        next(it)
    except StopIteration:
        pass
    else:
        raise ValueError(f"Malformed .auto tree: {auto!r}")

    return head


def parse_roots(
    auto: abc.Iterable[str] | Path | str,
    method: str,
    autox: bool = False,
) -> abc.Generator[HeadDependency | None, None, None]:
    """Find the root of each sentence in CCGbank's .auto format.

    Args:
        auto (abc.Iterable[str] | Path | str): The sentences to find the roots
            of or the path to a file containing the sentences.
        method (str): The head-finding method to use. See `get_head` for
            options.
        autox (bool, optional): Whether `auto` is in DepCCG's .autox format.
            Defaults to False.

    Yields:
        HeadDependency | None: The root of each given sentence.
    """
    with open(auto) if isinstance(auto, (Path, str)) else nullcontext(auto) as auto:
        for line in auto:
            line = line.strip()
            if line.startswith("ID="):
                continue
            elif not line:
                yield None
                continue
            yield parse_head(line, method, autox=autox)


head_methods = []
for name in list(globals()):
    if name.startswith("_head_"):
        head_methods.append(name[6:])



# This code in this file is provided under the C&C System License Agreement:
# C&C System Licence Agreement
#
# This Licence Agreement (the "Agreement") is entered, effective this
# date, by and between the "Licensor" (as defined below), and the
# individual executing this Agreement below as "Licensee" (hereinafter,
# the "Licensee").
#
# 1. Licensor hereby grants Licensee a royalty-free, non-exclusive,
# non-transferable licence to use the C&C system (herinafter referred to
# as the "Licensed Software") as follows solely for a Non-Commercial
# Purpose (as defined in paragraph 3 below):
#
#     (a) Licensee may prepare derivative works (the "Derivative Works")
#     which are based on or incorporate all or part of the Licensed
#     Software, including, without limitation, works (the "Adaptations")
#     which
#
#         (i) are translations of all or part of the Licensed Software
#         into different programming languages, or
#         (ii) are revisions, improvements or corrections to all or part
#         of the Licensed Software,
#
# provided that, Licensee shall treat all Derivative Works as Licensed
# Software under this Agreement; and
#
#     (b) Licensee may make only such copies of the Licensed Software as
#     are necessary for Licensee's development of the Derivative Works.
#
# 2. All copies of the Licensed Software and Derivative Works prepared
# in accordance with paragraph 1 shall retain the copyright notice
# appearing in the Licensed Software. If the Licensed Software includes
# computer programs in object code form, Licensee shall not de-compile,
# reverse engineer or disassemble such programs.
#
# 3. As used in this Agreement, "Non-Commercial Purpose" means use of
# the Licensed Software and Derivative Works solely for education or
# research.  "Non-Commercial Purpose" excludes, without limitation, any
# use of the Licensed Software or Derivative Works for, as part of, or
# in any way in connection with a product (including software) or
# service which is sold, offered for sale, licensed, leased, loaned or
# rented.
#
# 4. Licensee shall acknowledge use of the Licensed Software and
# Derivative Works in all publications of research based in whole or in
# part on their use through citation of the following publication:
#
# Parsing the WSJ using CCG and Log-Linear Models
# Stephen Clark and James R. Curran
# Proceedings of the 42nd Annual Meeting of the Association for
# Computational Linguistics (ACL-04), pp. 104-111, Barcelona, Spain 2004
#
# 5. Licensee hereby grants Licensor a non-exclusive, royalty-free, fully
# paid-up, worldwide, perpetual licence to:
#
#     (a) Reproduce, prepare derivative works based on and distribute all
#     or part of the Derivative Works; and
#
#     (b) Make, have made, use, offer to sell, sell, license or import any
#     products (including software) or services under any intellectual
#     property rights owned or licensed by Licensee which relate to
#
#         (i) all or part of the Derivative Works (including as executed
#         by a CPU), or
#         (ii) methods or concepts embodied in, or implemented through
#         the execution by a CPU of, the Derivative Works. Licensee
#         shall provide the contact person (or an alternative contact
#         notified to the Licensee from time to time)
#
# 	candc@it.usyd.edu.au
#
#         with feedback concerning Licensee's Derivative Works and, if
#         requested, provide such persons with source code copies of
#         Licensee's Derivative Works.
#
# 6. This Agreement is personal between Licensor and Licensee. No
# ownership interest in the Licensed Software (or the copy of which is
# provided by Licensor pursuant to paragraph 1) is transferred to
# Licensee. Licensee's interest in the Derivative Works is limited
# solely to Licensee's additions and the Derivative Works are subject in
# their entirety to Licensor's intellectual property rights. Licensor
# may assign or transfer to any company or person, or grant to any
# company or person a licence or sublicence under, all or part of its
# interest in any rights to the Licensed Software, this Agreement, or
# any licence granted to Licensor hereunder. Licensee may not assign or
# sublicense Licensee's rights hereunder.
#
# 7. Licensor may terminate this Agreement at any time by sending
# written notice of termination to Licensee at the address specified
# below.  Termination shall be effective as provided in the
# notice. Unless the notice shall provide otherwise, upon termination,
# Licensee shall destroy all copies of the Licensed Software and
# Derivative Works. Licensee's obligations under this Agreement,
# including any rights granted to Licensor pursuant to paragraph 5,
# shall survive and continue after termination.
#
# 8. Licensor has no obligation to support or maintain the Licensed
# Software and grants Licensee this right to use the Licensed Software
# "AS IS". LICENSEE ASSUMES TOTAL RESPONSIBILITY AND RISK FOR LICENSEE'S
# USE OF THE LICENSED SOFTWARE. LICENSOR DOES NOT MAKE, AND EXPRESSLY
# DISCLAIMS, ANY EXPRESS OR IMPLIED WARRANTIES, REPRESENTATIONS OR
# ENDORSEMENTS OF ANY KIND WHATSOEVER, INCLUDING, WITHOUT LIMITATION,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY, REASONABLE QUALITY OR
# FITNESS FOR A PARTICULAR PURPOSE, AND THE WARRANTIES OF TITLE OR
# NON-INFRINGEMENT. IN NO EVENT SHALL LICENSOR BE LIABLE FOR
#
#     (a) ANY INCIDENTAL, CONSEQUENTIAL, OR INDIRECT DAMAGES, DAMAGES
#     FOR LOSS OF PROFITS, BUSINESS INTERRUPTION, LOSS OF PROGRAMS OR
#     INFORMATION, AND ANY OTHER LOSS OR DAMAGE WHATSOEVER ARISING OUT
#     OF THE USE OF OR INABILITY TO USE THE SOFTWARE, EVEN IF LICENSOR
#     OR ANY OF ITS AUTHORIZED REPRESENTATIVES HAS BEEN ADVISED OF THE
#     POSSIBILITY OF SUCH DAMAGES,
#
#     (b) ANY CLAIM ATTRIBUTABLE TO ERRORS, OMISSIONS, OR OTHER
#     INACCURACIES IN THE SOFTWARE, OR
#
#     (c) ANY CLAIM BY ANY THIRD PARTY.
#
# 9. This Agreement shall be governed by and construed in accordance
# with the laws of England and the English courts shall have exclusive
# jurisdiction.
#
# LICENSOR
# University of Edinburgh
# James R. Curran
# University of Oxford
# University of Sydney

# The contents of this file come from the `evaluate` script included in the
# original C&C package as well as the `evaluate_new` script included in the Java
# C&C package. The code included here has been reproduced with permission for
# inclusion in the `catgram` package.

# These ignored deps are for the markedup file that comes with Java C&C
_ignore_deps_java = set(
    tuple(x.split())
    for x in r"""
rule_id 7
rule_id 11
rule_id 12
rule_id 14
rule_id 15
rule_id 16
rule_id 17
rule_id 51
rule_id 52
rule_id 56
rule_id 91
rule_id 92
rule_id 95
rule_id 96
rule_id 98
conj 1 0
((S[to]{_}\NP{Y}<1>){_}/(S[b]{Z}\NP{Y}){Z}<2>){_} 1 0
((S[to]{_}\NP{Y}<1>){_}/(S[b]{Z}\NP{Y}){Z}<2>){_} 1 2
((S[to]{_}\NP{Y}<1>){_}/(S[b]{Z}\NP{Y}){Z}<2>){_} 1 3
((S[to]{_}\NP{Y}<1>){_}/(S[b]{Z}\NP{Y}){Z}<2>){_} 1 6
((S[to]{_}\NP{Y}<1>){_}/(S[b]{Z}\NP{Y}){Z}<2>){_} 1 9
((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 6
((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_} 1 6
(((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_}/NP{W}<3>){_} 1 6
(S[X]{Y}/S[X]{Y}<1>){_} 1 13
(S[X]{Y}/S[X]{Y}<1>){_} 1 5
(S[X]{Y}/S[X]{Y}<1>){_} 1 55
((S[X]{Y}/S[X]{Y}){Z}\(S[X]{Y}/S[X]{Y}){Z}<1>){_} 2 97
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 4
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 93
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 8
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 94
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 18
been ((S[pt]{_}\NP{Y}<1>){_}/(S[ng]{Z}<2>\NP{Y}){Z}){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[pss]{Z}\NP{Y}){Z}<2>){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[adj]{Z}\NP{Y}){Z}<2>){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}\NP{Y}){Z}<2>){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[pt]{Z}\NP{Y}){Z}<2>){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[adj]{Z}\NP{Y}){Z}<2>){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[ng]{Z}\NP{Y}){Z}<2>){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y}){Z}){_} 1 0
going ((S[ng]{_}\NP{Y}<1>){_}/(S[to]{Z}<2>\NP{Y}){Z}){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[to]{Z}\NP{Y}){Z}<2>){_} 1 0
Here (S[adj]{_}\NP{Y}<1>){_} 1 0
# this is a dependency Julia doesn't have but looks okay
from (((NP{Y}\NP{Y}<1>){_}/(NP{Z}\NP{Z}){W}<3>){_}/NP{V}<2>){_} 1 0
""".strip().split(
        "\n"
    )
    if not x.startswith("#")
)

# And here are the original C&C ones, so that outputs from original C&C's
# `generate` program can be evaluated as well. Note the different format: here
# long-distance labels are indicated with an asterisk, and some of the slots are
# slightly different.
_ignore_deps_orig = set(
    tuple(x.split())
    for x in r"""
rule_id 7
rule_id 11
rule_id 12
rule_id 14
rule_id 15
rule_id 16
rule_id 17
rule_id 51
rule_id 52
rule_id 56
rule_id 91
rule_id 92
rule_id 95
rule_id 96
rule_id 98
conj 1 0
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 0
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 2
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 3
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 6
((S[to]{_}\NP{Z}<1>){_}/(S[b]{Y}<2>\NP{Z*}){Y}){_} 1 9
((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 6
((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_} 1 6
(((S[b]{_}\NP{Y}<1>){_}/PP{Z}<2>){_}/NP{W}<3>){_} 1 6
(S[X]{Y}/S[X]{Y}<1>){_} 1 13
(S[X]{Y}/S[X]{Y}<1>){_} 1 5
(S[X]{Y}/S[X]{Y}<1>){_} 1 55
((S[X]{Y}/S[X]{Y}){Z}\(S[X]{Y}/S[X]{Y}){Z}<1>){_} 2 97
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 4
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 93
((S[X]{Y}\NP{Z}){Y}\(S[X]{Y}<1>\NP{Z}){Y}){_} 2 8
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 94
((S[X]{Y}\NP{Z}){Y}/(S[X]{Y}<1>\NP{Z}){Y}){_} 2 18
been ((S[pt]{_}\NP{Y}<1>){_}/(S[ng]{Z}<2>\NP{Y*}){Z}){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
been ((S[pt]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 there 0
be ((S[b]{_}\NP{Y}<1>){_}/NP{Z}<2>){_} 1 There 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
been ((S[pt]{_}\NP{Y}<1>){_}/(S[adj]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[pt]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[adj]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[ng]{Z}<2>\NP{Y*}){Z}){_} 1 0
be ((S[b]{_}\NP{Y}<1>){_}/(S[pss]{Z}<2>\NP{Y*}){Z}){_} 1 0
going ((S[ng]{_}\NP{Y}<1>){_}/(S[to]{Z}<2>\NP{Y*}){Z}){_} 1 0
have ((S[b]{_}\NP{Y}<1>){_}/(S[to]{Z}<2>\NP{Y*}){Z}){_} 1 0
Here (S[adj]{_}\NP{Y}<1>){_} 1 0
# this is a dependency Julia doesn't have but looks okay
from (((NP{Y}\NP{Y}<1>){_}/(NP{Z}\NP{Z}){W}<3>){_}/NP{V}<2>){_} 1 0
""".strip().split(
        "\n"
    )
    if not x.startswith("#")
)

deps_to_ignore = _ignore_deps_java | _ignore_deps_orig


def ignore_dep(pred: str, cat: str, slot: str, arg: str, rule_id: str):
    # TODO: document
    return (
        ("rule_id", rule_id) in deps_to_ignore
        or (cat, slot, rule_id) in deps_to_ignore
        or (pred, cat, slot, rule_id) in deps_to_ignore
        or (pred, cat, slot, arg, rule_id) in deps_to_ignore
    )



from collections.abc import Mapping, Sequence
from functools import cache
import re
from typing import NamedTuple

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing import TypeAlias

    Self: TypeAlias = "LexicalDecomposition"

from lambda_calculus import terms

from catgram import CategoryTree


class ProofFrameNode(NamedTuple):
    """
    A node in a proof frame.

    * The `atom` field is the (atomic) category label for the node (e.g.,
      `'S'`, `'NP'`, `'N'`, etc.).
    * The `positive` field indicates node polarity (`true` for positive,
      `false` for negative).
    * The `children` field is a tuple of integers indicating the *relative*
      positions of the node's children.

    This type is a tuple; immutability allows instances to be used as values in
    a cache such that lexical categories can re-use nodes and construct new
    ones without altering nodes used for other categories.
    """

    atom: str
    positive: bool
    children: tuple[int, ...] = tuple()

    def __repr__(self):
        return f"{self.atom}{'+' if self.positive else '-'}{self.children}"


class LexicalDecomposition(NamedTuple):
    """
    The result of decomposing a polarized lexical category.

    * The `nodes` field is a tuple of `ProofFrameNode`s, storing them in the
      order indicated by the decomposition rules.
    * The `edges` field is a tuple of integer 2-tuples, indicating the source
      and destination indices of the nodes (as in the `nodes` tuple) for each
      edge (regular or Lambek) as constructed by the lexical decomposition.
    * The `target` field is the index of the node in the `nodes` tuple
      corresponding to the "target" atom of the lexical category. The target
      node is the root node of the lexical decomposition. In constructing
      semantic terms in term graphs, token labels are assigned to target atoms.
    """

    nodes: tuple[ProofFrameNode, ...]
    edges: tuple[tuple[int, int], ...]
    target: int

    @classmethod
    @cache
    def decompose_category(cls, category: CategoryTree, positive: bool = False) -> Self:
        """
        Decompose a polarized lexical category according to the lexical
        decomposition rules.

        Args:
            category (CategoryTree): The lexical category to decompose.
            positive (bool, optional): The polarity of the category. Defaults to
                `False`.

        Returns:
            LexicalDecomposition: The decomposition result.

        Raises:
            ValueError: If the category is not atomic and doesn't have a
                forward or backward slash as its root or if `positive` is
                something other than `True` or `False`.
        """
        if category.is_atomic:
            return cls((ProofFrameNode(category.root, positive),), tuple(), 0)
        result = cls.decompose_category(category.result, positive)
        argument = cls.decompose_category(category.argument, not positive)
        target = result.nodes[result.target]

        match category.root, positive:
            # This is a bit of a mess, but tracking the children here makes the
            # traversal in TermGraph._get_term() much cleaner
            case ("/", False) | ("\\", True):
                shift = len(result.nodes)
                edge = (result.target, shift + argument.target)
                nodes = list(result.nodes)
                nodes[result.target] = nodes[result.target]._replace(
                    children=(edge[1] - edge[0],) + target.children
                )
                left = result._replace(nodes=tuple(nodes))
                right = argument
                # left, right = result, argument
            case ("/", True) | ("\\", False):
                shift = len(argument.nodes)
                edge = (shift + result.target, argument.target)
                left = argument
                nodes = list(result.nodes)
                nodes[result.target] = nodes[result.target]._replace(
                    children=(edge[1] - edge[0],) + target.children
                )
                right = result._replace(nodes=tuple(nodes))
                # left, right = argument, result
            case _:
                raise ValueError(f"invalid category: {category}")

        nodes = left.nodes + right.nodes
        edges = [
            edge,
            *left.edges,
            *[(src + shift, dst + shift) for src, dst in right.edges],
        ]

        return cls(nodes, tuple(edges), edge[0])


class TermGraph:
    """
    A representation of the term graph of a sequent in the Lambek calculus.
    See Fowler (2016) for details.
    """

    # lambda_calculus doesn't like these letters, but doesn't put the list
    # anywhere (it just shows up as a literal in the code)
    _unsafe = re.compile(r"[().λ]")

    def __init__(
        self,
        categories: Sequence[CategoryTree | str | None],
        *,
        sentential: int | None = 0,
        linkage: Mapping[int, int] | None = None,
        words: Sequence | None = None,
    ):
        """
        Construct a term graph for the given sequent and linkage with optional
        token labels.

        Args:
            categories: A `Sequence` of any combination of `CategoryTree`s,
                `str`s, and `None`s. Strings are converted to `CategoryTree`s
                and `None` categories can be useful for representing words with
                no lexical category, a convention sometimes used for some
                punctuation (e.g., in LCGbank).
            sentential (int | None): The index of the sentential category.
                Can be negative as per the usual Python negative-index
                convention. Defaults to `0`. `None` indicates that there is no
                sentential category.
            linkage: A `Mapping` of source (positive) atom indices to
                destination (negative) atom indices indicating links in the
                term graph.
            words: A `Sequence` of labels used to label lexical categories'
                target atoms.

        Raises:
            ValueError: If `words` is not None and has a different length than
                `categories`.
        """
        self.categories = [
            CategoryTree.from_str(cat) if isinstance(cat, str) else cat
            for cat in categories
        ]
        if sentential is not None:
            sentential = sentential % len(self.categories)
        self.linkage = {} if linkage is None else linkage
        self.words = [f"<w_{i}>" for i in range(len(categories))]
        if words is not None:
            if len(words) != len(categories):
                raise ValueError("number of words must match number of categories")
            for i, word in enumerate(words):
                word = self._unsafe.sub("", word)
                if word:
                    self.words[i] = f"{word}_{i}"

        self.decomps = []
        for i, cat in enumerate(self.categories):
            if cat is None:
                self.decomps.append(None)
                continue
            else:
                self.decomps.append(
                    LexicalDecomposition.decompose_category(cat, i == sentential)
                )
        if sentential is not None:
            self.decomps[sentential] = LexicalDecomposition.decompose_category(
                self.categories[sentential], True
            )
        self.nodes = []
        self.clens = []
        for decomp in self.decomps:
            self.clens.append(len(self.nodes))
            if decomp is None:
                continue
            self.nodes.extend(decomp.nodes)

        self.labels = [f"{node.atom}_{idx}" for idx, node in enumerate(self.nodes)]
        for word, decomp, clen in zip(
            self.words, self.decomps, self.clens, strict=True
        ):
            if decomp is None:
                continue
            self.labels[clen + decomp.target] = word

    def get_term(self, idx: int = 0) -> terms.Term:
        """
        Get the semantic term for the token at the given index.

        Args:
            idx (int): The index for the token.

        Returns:
            Term: the semantic term for the token as situated in the term graph.
                If the index provided is that of the sentential category and
                `self` is a valid term graph, this will return the semantic
                term for the entire sentence.
        """
        return self._get_term(self.clens[idx] + self.decomps[idx].target)

    def _get_term(self, idx: int) -> terms.Term:
        if (node := self.nodes[idx]).positive:
            try:
                term = self._get_term(self.linkage[idx])
            except KeyError:
                term = terms.Variable.with_valid_name(self.labels[idx])
            for child in node.children:
                term = terms.Abstraction(self.labels[idx + child], term)
        else:
            term = terms.Variable.with_valid_name(self.labels[idx])
            for child in node.children:
                term = terms.Application(term, self._get_term(idx + child))

        return term



from dataclasses import dataclass
from functools import cache, cached_property
import logging
import re
from typing import ClassVar, overload, Optional

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing import TypeAlias

    Self: TypeAlias = "CategoryTree"


@dataclass(frozen=True, repr=False)
class CategoryTree:
    """A tree representation of a category."""

    atom_re: ClassVar[re.Pattern] = re.compile(r"^[^\s/\\()]+$")
    atom_delim: ClassVar[re.Pattern] = re.compile(r"([()\\/])")

    root: str
    result: Optional[Self] = None
    argument: Optional[Self] = None

    def __post_init__(self: Self):
        # validate the fields' values
        if not isinstance(self.root, str):
            raise TypeError(f"root must be {str} instance; got {self.root!r}")
        if self.result is None:
            if self.argument is not None:
                raise ValueError("argument without result")
            if not self.atom_re.fullmatch(self.root):
                raise ValueError(
                    f"atomic category must not have spaces, slashes, "
                    f"backslashes, or parentheses; got {self.root!r}"
                )
        else:
            if self.argument is None:
                raise ValueError("result without argument")
            for _child in ("result", "argument"):
                if not isinstance(child := getattr(self, _child), cls := type(self)):
                    raise TypeError(
                        f"{_child} must be a {cls.__name__} instance; got {child!r}"
                    )
            if self.root not in {"/", "\\"}:
                raise ValueError(
                    f"root for complex category must be '/' or '\\'; "
                    f"got {self.root!r}"
                )

    @classmethod
    def from_str(
        cls,
        cat: str,
        left_associative: bool = False,
        lambek: bool = False,
        strict: bool = True,
    ) -> Self:
        """
        Convert a string representation of a category to a CategoryTree.

        Args:
            cat (str): The category in string form.
            left_associative (bool, optional): Whether the category is
                left-associative (alternative is full parenthesization).
                Defaults to `False` (full parenthesization).
            lambek (bool, optional): Whether the category is in Lambek order
                (alternative is Steedman order). Defaults to `False` (
                Steedman order).
            strict (bool, optional): Whether to be strict when parsing the
                string (alternative is to make some simple assumptions to try
                to parse certain malformed categories). Defaults to `True`.
                `False` can be useful for categories predicted by a
                constructive supertagger if it might have minor malformations
                that can be straightforwardly corrected.

        Returns:
            CategorTree: The CategoryTree representation of the input category.
        """
        if (key := (cat, lambek)) in INSTANCES:
            return INSTANCES[key]
        cat_ = tuple(a for a in cls.atom_delim.split(cat) if a)
        return cls._parse_str(cat_, left_associative, lambek, strict)[0]

    @classmethod
    def _parse_str(
        cls,
        cat: tuple[str],
        l_assoc: bool,
        lambek: bool,
        strict: bool,
        idx: int = 0,
        in_paren: bool = False,
    ) -> tuple[Self, int]:
        root = left = right = None
        i = idx
        while i < len(cat):
            if not isinstance(cat[i], str):
                raise TypeError(
                    f"category must be str; got {cat[i]!r} at " f"index {i}"
                )
            if cat[i] == ")":
                if not in_paren:
                    msg = (
                        f'xtraneous closing parenthesis in "{"".join(cat)}" '
                        f"at index {i}"
                    )
                    if strict:
                        raise MalformedCategoryError(f"E{msg}")
                    logging.warning(f"Ignoring e{msg}")
                break
            if right is not None:
                if not l_assoc or cat[i] not in "/\\":
                    # only valid case for iterating after having read a right
                    # node is reading a closing paren, which was handled above
                    if not in_paren or strict:
                        raise MalformedCategoryError(
                            f'Extraneous stuff "{cat[i]}" in "{"".join(cat)}" '
                            f"at index {i} after right node "
                            f'"{str(right)}": {cat[i]}'
                        )
                    # if being lenient (not strict), allow some leeway if in
                    # parenthetical
                    logging.warning(
                        f"Assuming missing closing parenthesis in "
                        f'"{"".join(cat)}" at index {i}'
                    )
                    i -= 1  # unconsume this token
                    break
            if cat[i] in "/\\":
                c_str = "".join(cat[idx:i])
                if root is None:
                    raise MalformedCategoryError(
                        f'Missing left node in "{"".join(cat)}" '
                        f'before "{cat[i]}" at index {i}'
                    )
                if left is None:
                    if isinstance(root, cls):
                        left = root
                    else:
                        if (c_str, lambek) in INSTANCES:
                            left = INSTANCES[c_str, lambek]
                            # TODO: nix sanity check after adding some tests
                            assert left == cls(root)
                        else:
                            INSTANCES[c_str, lambek] = left = cls(root)
                elif not l_assoc or right is None:
                    # TODO: I suspect this case should be "and", not "or", but
                    #  at least this way it's overzealous rather than lax
                    raise MalformedCategoryError(
                        f'Two consecutive slashes in "{"".join(cat)}" at index "{i}"'
                    )
                else:
                    # build a tree from existing root, left, right and assign
                    # to left
                    if (key := (c_str, lambek)) in INSTANCES:
                        left = INSTANCES[key]
                        # TODO: nix sanity check after adding some tests
                        if lambek and root == "\\":
                            assert left == cls(root, right, left)
                        else:
                            assert left == cls(root, left, right)
                    elif lambek and root == "\\":
                        INSTANCES[key] = left = cls(root, right, left)
                    else:
                        INSTANCES[key] = left = cls(root, left, right)
                    right = None
                root = cat[i]
            else:
                p = cat[i]
                if p == "(":
                    p, i = cls._parse_str(
                        cat, l_assoc, lambek, strict, idx=i + 1, in_paren=True
                    )

                if left is None:
                    if root is not None:
                        raise MalformedCategoryError(
                            f'Extraneous stuff in "{"".join(cat)}" at index '
                            f'{i} after "{root}"; expected slash'
                        )
                    if l_assoc and isinstance(p, cls):
                        raise MalformedCategoryError(
                            "Parenthesized expression in left-associative "
                            f'mode in "{"".join(cat)}" at index {i} before slash'
                        )
                    root = p
                else:
                    if not isinstance(p, cls):
                        if (key := (p, lambek)) in INSTANCES:
                            p = INSTANCES[key]
                            # TODO: nix sanity check after adding some tests
                            assert p == cls(key[0])
                        else:
                            INSTANCES[key] = p = cls(p)
                    right = p
            i += 1
        else:
            if in_paren:
                msg = f'nclosed parenthetical in "{"".join(cat)}" at index {i}'
                if strict:
                    raise MalformedCategoryError(f"U{msg}")
                logging.warning(f"Ignoring u{msg}")
        if root is None:
            raise MalformedCategoryError("Empty category")

        cat_ = "".join(cat[idx:i])
        if isinstance(root, cls):
            logging.warning(f'Stripping redundant parentheses in "{cat_}"')
            cat_ = cat_[1:-1]
            # TODO: nix sanity check after adding some tests
            assert root == INSTANCES[cat_, lambek]
            return root, i
        if root in "/\\" and right is None:
            raise MalformedCategoryError(
                f'Missing right node after "{root}" at index {i} in "{cat_}"'
            )

        key = cat_, lambek
        if key in INSTANCES:
            cat_ = INSTANCES[key]
            # TODO: nix sanity check after adding some tests
            if lambek and root == "\\":
                assert cat_ == cls(root, right, left)
            else:
                assert cat_ == cls(root, left, right)
        elif lambek and root == "\\":
            INSTANCES[key] = cat_ = cls(root, right, left)
        else:
            INSTANCES[key] = cat_ = cls(root, left, right)

        return cat_, i

    @cached_property
    def is_complex(self: Self) -> bool:
        """Whether this category is complex (i.e., not atomic)."""
        return not self.is_atomic

    @cached_property
    def is_atomic(self: Self) -> bool:
        """Whether this category is atomic (i.e., takes no arguments)."""
        return self.result is None

    @cached_property
    def target(self: Self) -> str | None:
        """
        The target atomic category of a complex category (i.e., the result of
        the result of the result ... of the category until the result is
        atomic).
        """
        if self.is_atomic:
            return None
        if self.result.is_atomic:
            return self.result.root
        return self.result.target

    @cached_property
    def arity(self: Self) -> int:
        """
        The number of arguments this category takes before yielding its target
        (atomic) category.
        """
        return 0 if self.is_atomic else self.result.arity + 1

    @cached_property
    def _len(self: Self) -> int:
        """
        The number of atomic categories in this category (i.e., number of leaf
        nodes in tree).
        """
        if self.is_atomic:
            return 1
        return len(self.result) + len(self.argument)

    def __len__(self: Self) -> int:
        """
        The number of atomic categories in this category (i.e., number of leaf
        nodes in tree).
        """
        return self._len

    @cached_property
    def degree(self: Self) -> int:
        """Category tree height plus one, as defined by Moot and Retoré (2012)."""
        if self.is_atomic:
            return 1  # would be 0 for binary tree depth
        return max(self.result.degree, self.argument.degree) + 1

    @cached_property
    def order(self: Self) -> int:
        """A measure of highest "complexity" of this category's arguments."""
        if self.is_atomic:
            return 0
        return max(self.result.order, self.argument.order + 1)

    @cached_property
    def size(self: Self) -> int:
        """Number of nodes in this category tree."""
        if self.is_atomic:
            return 1
        return self.result.size + self.argument.size + 1

    def __str__(self: Self) -> str:
        """
        Fully-parenthesized string representation of this category in Steedman
        order.
        """
        return self.to_str()

    @overload
    def to_str(
        self: Self,
        *,
        pol_start: None = ...,
        positive: bool = ...,
        l_assoc: bool = ...,
        lambek: bool = ...,
    ) -> str:
        ...

    @overload
    def to_str(
        self: Self,
        *,
        pol_start: int,
        positive: bool = ...,
        l_assoc: bool = ...,
        lambek: bool = ...,
    ) -> tuple[str, int]:
        ...

    def to_str(
        self: Self,
        *,
        pol_start: int | None = None,
        positive: bool = False,
        l_assoc: bool = False,
        lambek: bool = False,
    ) -> str | tuple[str, int]:
        """String representation of this category

        Args:
            pol_start: If not None, add an index to each atomic category
                that indicates the position of that (polarized) primitive in
                the lexical decomposition of the category. The integer value of
                this parameter specifies the offset to use for the first index.
            positive: Whether the category has positive polarity. Has no effect
                if `pol_start_idx` is None.
            l_assoc: Whether to use left-associative notation
            lambek: Whether to use Lambek notation
        Returns:
            If `pol_start_idx` is None, returns a string representation of this
            category according to the specified notation. If `pol_start_idx` is
            not None, returns a 2-tuple: the first element is the string
            representation of this category and the second element is the final
            atomic index used for this category.
        """
        seq_, pol_idx = self._inorder(l_assoc, lambek, positive)
        if pol_start is None:
            return "".join(seq_)

        seq = list(seq_)
        leaf_idxs = [i for i, s in enumerate(seq) if not self.atom_delim.fullmatch(s)]
        # TODO: nix sanity check after adding some tests
        assert len(leaf_idxs) == len(self) == len(pol_idx)
        # for i, l in enumerate(leaf_idxs, start=pol_start):
        #    seq[l] = f"{seq[l]}_{i}"
        for i, j in zip(leaf_idxs, pol_idx, strict=True):
            seq[i] = f"{seq_[i]}_{pol_start + j}"
        return "".join(seq), pol_start + len(leaf_idxs)

    def __repr__(self: Self) -> str:
        # cls = self.__class__.__name__
        # if self.is_atomic:
        #    return f"{cls}({self.root})"
        # return (
        #     f"{cls}(result={self.result!r}, root={self.root}, "
        #     f"argument={self.argument!r})"
        # )
        # The above can get pretty long for larger categories, so just use the
        # string representation wrapped in the class name.
        return f"{self.__class__.__name__}({self.to_str()})"

    @cache
    def _inorder(
        self: Self, l_assoc: bool, lambek: bool, positive: bool
    ) -> tuple[tuple[str, ...], tuple[int, ...]]:
        if self.is_atomic:
            return (self.root,), (0,)
        result, result_idxs = self.result._inorder(l_assoc, lambek, positive)
        argument, argument_idxs = self.argument._inorder(l_assoc, lambek, not positive)
        if not l_assoc and len(result) > 1:
            result = ("(", *result, ")")
        if len(argument) > 1:
            argument = ("(", *argument, ")")

        if (self.root == "/") == positive:
            offset = len(argument_idxs)
            result_idxs = tuple(i + offset for i in result_idxs)
        else:
            offset = len(result_idxs)
            argument_idxs = tuple(i + offset for i in argument_idxs)

        if lambek and self.root == "\\":
            return argument + (self.root,) + result, argument_idxs + result_idxs
        return result + (self.root,) + argument, result_idxs + argument_idxs

    @cached_property
    def func_seq(self: Self) -> tuple[str, ...]:
        """
        The functorial sequence that this category decomposes into, as defined
        by Bhargava and Penn (ACL 2023).
        """
        if self.is_atomic:
            return (self.root,)
        argument_str = self.argument.to_str()
        if self.argument.is_atomic:
            argument_str = f"{self.root}{argument_str}"
        else:
            argument_str = f"{self.root}({argument_str})"
        return self.result.func_seq + (argument_str,)

    @cache
    def get_arg(self: Self, slot: int) -> Self:
        if slot > self.arity:
            raise ValueError(
                f"invalid slot {slot!r} for category {self} with arity {self.arity}"
            )
        if slot == self.arity:
            return self.argument if slot else self
        return self.result.get_arg(slot)


class MalformedCategoryError(Exception):
    pass


INSTANCES: dict[tuple[str, bool], CategoryTree] = {}
